use std::fmt::{Debug, Formatter};
use std::sync::Arc;
use tokio::fs;
use tokio_stream::wrappers::ReceiverStream;
use tonic::{metadata::Ascii, metadata::MetadataValue, Request, Response, Status};
use tracing::{field, info, instrument};

use hg::changelog::Changelog;
use hg::revlog::{NodePrefix, RevlogError, NULL_REVISION};

use crate::config::Config;
use crate::gitaly::commit_service_server::{CommitService, CommitServiceServer};
use crate::gitaly::{
    FindCommitRequest, FindCommitResponse, ListCommitsByOidRequest, ListCommitsByOidResponse,
    Repository,
};
use crate::gitlab::revision::gitlab_revision_node_prefix;
use crate::message;
use crate::metadata::correlation_id;
use crate::repository::{
    default_repo_spec_error_status, load_repo_and_stream, load_repo_and_then, repo_path,
    RequestWithRepo,
};
use crate::streaming::ResultResponseStream;

#[derive(Debug)]
pub struct CommitServiceImpl {
    config: Arc<Config>,
}

const COMMITS_CHUNK_SIZE: usize = 50;

/// Return the [`ListCommitsByOidResponse`] for a chunk of oids (changeset ids)
///
/// Behaving like Gitaly imposes that
/// - oids that cannot be found in the repository are ignored. This is actually one of the
///   use cases for the `ListCommitsByOid` call: it can be called indiscriminately with anything
///   that looks like a hash, in particular from Markdown renderings
/// - oids that are not even hexadecimal are ignored, too. This is possibly coincidental.
///
/// [`RevlogError`] is returned in cases of repository corruption.
fn commits_by_oid_chunk(
    cl: &Changelog,
    oids: &[String],
) -> Result<ListCommitsByOidResponse, RevlogError> {
    // Not using Vec::with_capacity(oids.len()), because clients will often call this with a single
    // argument, precisely to know whether it is a commit id prefix in this repo (happens a lot
    // in Markdown rendering), and this will result in needless allocations.
    let mut commits = Vec::new();
    for oid in oids {
        match NodePrefix::from_hex(oid) {
            Err(_) => {
                continue; // non-hexadecimal values for oid are just ignored by Gitaly
            }
            Ok(node_pref) => match cl.rev_from_node(node_pref) {
                Ok(NULL_REVISION) => (),
                Ok(rev) => commits.push(message::commit(cl.entry_for_rev(rev)?)?),
                Err(RevlogError::InvalidRevision) => (),
                Err(e) => {
                    return Err(e);
                }
            },
        }
    }
    Ok(ListCommitsByOidResponse { commits })
}

impl RequestWithRepo for ListCommitsByOidRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for FindCommitRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

#[tonic::async_trait]
impl CommitService for CommitServiceImpl {
    type ListCommitsByOidStream = ReceiverStream<Result<ListCommitsByOidResponse, Status>>;

    async fn list_commits_by_oid(
        &self,
        request: Request<ListCommitsByOidRequest>,
    ) -> Result<Response<Self::ListCommitsByOidStream>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_list_commits_by_oid(inner, correlation_id(&metadata))
            .await
    }

    async fn find_commit(
        &self,
        request: Request<FindCommitRequest>,
    ) -> Result<Response<FindCommitResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_find_commit(inner, correlation_id(&metadata))
            .await
    }
}

struct FindCommitTracingRequest<'a>(&'a FindCommitRequest);

impl Debug for FindCommitTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("FindCommitRequest")
            .field("repository", &self.0.repository)
            .field("revision", &String::from_utf8_lossy(&self.0.revision))
            .finish()
    }
}

impl CommitServiceImpl {
    #[instrument(name = "list_commits_by_oid", skip(self))]
    async fn inner_list_commits_by_oid(
        &self,
        request: ListCommitsByOidRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> ResultResponseStream<ListCommitsByOidResponse> {
        info!("Processing");

        load_repo_and_stream(
            self.config.clone(),
            request,
            default_repo_spec_error_status,
            move |req, repo, tx| match repo.changelog() {
                Ok(cl) => {
                    for oid_chunk in req.oid.chunks(COMMITS_CHUNK_SIZE) {
                        match commits_by_oid_chunk(&cl, oid_chunk) {
                            Ok(resp) => {
                                if !resp.commits.is_empty() {
                                    tx.send(Ok(resp));
                                }
                            }
                            Err(e) => tx.send(Err(Status::internal(format!(
                                "Repository corruption {:?}",
                                e
                            )))),
                        }
                    }
                }
                Err(e) => tx.send(Err(Status::internal(format!(
                    "Could not open repository changelog: {:?}",
                    e
                )))),
            },
        )
    }

    #[instrument(name="find_commit",
                 skip(self, request),
                 fields(request=field::debug(FindCommitTracingRequest(&request))))]
    async fn inner_find_commit(
        &self,
        request: FindCommitRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<FindCommitResponse>, Status> {
        info!("Processing");
        if request.revision.is_empty() {
            return Err(Status::invalid_argument("empty revision"));
        }

        if let Some(gl_repo) = &request.repository {
            let config = self.config.clone();
            let path = repo_path(&config, gl_repo).map_err(default_repo_spec_error_status)?;
            if !match fs::metadata(&path).await {
                Ok(md) => md.is_dir(),
                Err(_) => false,
            } {
                return Err(Status::not_found(format!(
                    "Mercurial repository at {} not found",
                    path.display()
                )));
            }

            let store_path = path.join(".hg/store");
            match gitlab_revision_node_prefix(&store_path, &request.revision)
                .await
                .map_err(|e| Status::internal(format!("Error resolving revision: {:?}", e)))?
            {
                None => {
                    info!("Revision not resolved");
                    return Ok(Response::new(FindCommitResponse::default()));
                }
                Some(node_prefix) => {
                    info!("Revision resolved as {:x}", &node_prefix);
                    let commit = load_repo_and_then(
                        self.config.clone(),
                        request,
                        default_repo_spec_error_status,
                        move |_req, repo| {
                            let cl = repo.changelog().map_err(|e| {
                                Status::internal(format!("Could not open changelog: {:?}", e))
                            })?;

                            match message::commit_for_node_prefix(&cl, node_prefix) {
                                // TODO discuss upstream: InvalidRevision is incorrect wording:
                                // it does represent both invalid input (working dir Node),
                                // and not found node prefixes. The latter case justifies our
                                // choice.
                                // Note that at this stage we cannot tell apart NodePrefix due to
                                // caller passing a hash from those read in state files (would be an
                                // inconsistency for them not to resolve).
                                Err(RevlogError::InvalidRevision) => Ok(None),
                                Err(e) => {
                                    Err(Status::internal(format!("Repository corruption {:?}", e)))
                                }
                                Ok(r) => Ok(r),
                            }
                        },
                    )
                    .await?;

                    return Ok(Response::new(FindCommitResponse { commit }));
                }
            };
        }
        Err(Status::invalid_argument(
            "GetStorageByName: no such storage: \"\"",
        ))
    }
}

/// Takes care of boilerplate that would instead be in the startup sequence.
pub fn commit_server(config: &Arc<Config>) -> CommitServiceServer<CommitServiceImpl> {
    CommitServiceServer::new(CommitServiceImpl {
        config: config.clone(),
    })
}
