// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::fmt::{Debug, Formatter};
use std::string::String;
use std::sync::Arc;
use tonic::{metadata::Ascii, metadata::MetadataValue, Request, Response, Status};
use tracing::{field, info, instrument};

use crate::config::Config;
use crate::gitaly::ref_service_server::{RefService, RefServiceServer};
use crate::gitaly::{RefExistsRequest, RefExistsResponse};
use crate::gitlab::revision::{map_full_ref, RefError};
use crate::metadata::correlation_id;
use crate::repository::{default_repo_spec_error_status, repo_path};

#[derive(Debug)]
pub struct RefServiceImpl {
    config: Arc<Config>,
}

#[tonic::async_trait]
impl RefService for RefServiceImpl {
    async fn ref_exists(
        &self,
        request: Request<RefExistsRequest>,
    ) -> Result<Response<RefExistsResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_ref_exists(inner, correlation_id(&metadata))
            .await
    }
}

struct RefExistsTracingRequest<'a>(&'a RefExistsRequest);

impl Debug for RefExistsTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("RefExistRequest")
            .field("repository", &self.0.repository)
            .field("ref", &String::from_utf8_lossy(&self.0.r#ref))
            .finish()
    }
}

impl RefServiceImpl {
    #[instrument(name="ref_exists",
                 skip(self, request),
                 fields(request=field::debug(RefExistsTracingRequest(&request))))]
    async fn inner_ref_exists(
        &self,
        request: RefExistsRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<RefExistsResponse>, Status> {
        info!("Processing");

        if let Some(gl_repo) = &request.repository {
            let repo_path =
                repo_path(&self.config, gl_repo).map_err(default_repo_spec_error_status)?;
            let store_path = repo_path.join(".hg/store");

            let value = match map_full_ref(&store_path, &request.r#ref, |_tr| (), |_ka| ()).await {
                Ok(()) => Ok(true),
                Err(RefError::NotFound) => Ok(false),
                Err(RefError::MissingRefName) => Ok(false),
                Err(RefError::NotAFullRef) => Err(Status::invalid_argument("invalid refname")),
                Err(RefError::GitLabStateFileError(e)) => Err(Status::internal(format!("{:?}", e))),
            }?;

            return Ok(Response::new(RefExistsResponse { value }));
        }
        Err(Status::invalid_argument(
            "GetStorageByName: no such storage: \"\"",
        ))
    }
}

/// Takes care of boilerplate that would instead be in the startup sequence.
pub fn ref_server(config: &Arc<Config>) -> RefServiceServer<RefServiceImpl> {
    RefServiceServer::new(RefServiceImpl {
        config: config.clone(),
    })
}
