// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::sync::Arc;
use tokio::fs;
use tonic::{metadata::Ascii, metadata::MetadataValue, Request, Response, Status};
use tracing::{info, instrument};

use crate::config::Config;
use crate::gitaly::repository_service_server::{RepositoryService, RepositoryServiceServer};
use crate::gitaly::{RepositoryExistsRequest, RepositoryExistsResponse};
use crate::metadata::correlation_id;
use crate::repository::{default_repo_spec_error_status, repo_path};

#[derive(Debug)]
pub struct RepositoryServiceImpl {
    config: Arc<Config>,
}

#[tonic::async_trait]
impl RepositoryService for RepositoryServiceImpl {
    async fn repository_exists(
        &self,
        request: Request<RepositoryExistsRequest>,
    ) -> Result<Response<RepositoryExistsResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_repository_exists(inner, correlation_id(&metadata))
            .await
    }
}

impl RepositoryServiceImpl {
    #[instrument(name = "repository_exists", skip(self))]
    async fn inner_repository_exists(
        &self,
        request: RepositoryExistsRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<RepositoryExistsResponse>, Status> {
        info!("Processing");

        if let Some(gl_repo) = &request.repository {
            let path = repo_path(&self.config, gl_repo).map_err(default_repo_spec_error_status)?;
            let res = match fs::metadata(path).await {
                Ok(md) => md.is_dir(),
                Err(_) => false,
            };
            return Ok(Response::new(RepositoryExistsResponse { exists: res }));
        }
        Err(Status::invalid_argument(
            "GetStorageByName: no such storage: \"\"",
        ))
    }
}

/// Takes care of boilerplate that would instead be in the startup sequence.
pub fn repository_server(config: &Arc<Config>) -> RepositoryServiceServer<RepositoryServiceImpl> {
    RepositoryServiceServer::new(RepositoryServiceImpl {
        config: config.clone(),
    })
}
